; --------------------------------------------------------------------------
; AVR Thermometer
; (c) 2003 Andreas Schroeder <andreas@a-netz.de>
;
; Module: Display routines
;
; 4-bit interface used, all pins connected to port b
;
;  function | rs | rw | e | - | d7 | d6 | d5 | d4
;  bit      | 7  | 6  | 5 | 4 |  3 | 2  | 1  | 0
;
; --------------------------------------------------------------------------
.EQU	DISP_PORT = PORTB

.EQU    DISP_BIT_ENABLE = 5
.EQU    DISP_BIT_RW = 6
.EQU    DISP_BIT_RS = 7

.EQU    DISP_ADDR_FIRST_ROW = 0x80
.EQU    DISP_ADDR_SECOND_ROW = 0xC0
.EQU	DISP_CMD_CLEAR = 0x01

; set the display status (cursor, cursor blink, display on / off)
.EQU	DISP_CMD_STATUS = 0x08
.EQU	DISP_FLAG_CURSOR_MODE = 0
.EQU	DISP_FLAG_CURSOR = 1
.EQU	DISP_FLAG_ONOFF = 2

.EQU	DISP_PULSE_DELAY = 0x02
.EQU	DISP_SHORT_DELAY = 0x10

; --------------------------------------------------------------------------
; Initialization of the display interface
; --------------------------------------------------------------------------
.MACRO disp_init_interface
	; wait apx. 32 ms to let the display make
	; its internal init
	ldi		RP1, 8
	ldi		RP0, 0
	rcall	timer_start16

	; save Acc
	push	Acc

	ldi		Acc, 0x03
	out		DISP_PORT, Acc
	disp_pulse_enable

	; wait 8 ms
	ldi		RP1, 2
	ldi		RP0, 0
	rcall	timer_start16

	disp_pulse_enable

    ; interface to 4 bit
	disp_write
	ldi     Acc, 0x02
	out     DISP_PORT, Acc
	disp_pulse_enable

	pop		Acc
.ENDMACRO

.MACRO disp_enable
	sbi     DISP_PORT, DISP_BIT_ENABLE
.ENDMACRO

.MACRO disp_disable
	cbi     DISP_PORT, DISP_BIT_ENABLE
.ENDMACRO

.MACRO disp_data
	sbi     DISP_PORT, DISP_BIT_RS
.ENDMACRO

.MACRO disp_instruction
	cbi     DISP_PORT, DISP_BIT_RS
.ENDMACRO

.MACRO disp_pulse_enable
	ldi		RP0, DISP_PULSE_DELAY
			disp_enable
	rcall   timer_start8
			disp_disable
	rcall	timer_start8
.ENDMACRO

.MACRO disp_read
	push	Acc
	; low nibble as input
	ldi     Acc, 0xf0
	out     DDRB, Acc

	; set low nibble to 0 (high z)
	in      Acc, DISP_PORT

	; clear data bits
	andi    Acc, 0b11000000

	; set read bit
	ori     Acc, 0b01000000
	out     DISP_PORT, Acc
	pop		Acc
.ENDMACRO

.MACRO disp_write
	push	Acc
	; all lines to output
	ldi     Acc, 0xff
	out     DDRB, Acc

	in      Acc, DISP_PORT
	; read bit cleared
	andi    Acc, 0b10000000
	out     DISP_PORT, Acc
	pop		Acc
.ENDMACRO

; --------------------------------------------------------------------------
; Proc: Write a byte to the display
;
; Params: RP0 - the byte to write
; --------------------------------------------------------------------------
disp_put_byte:
	push	Acc
	push	Acc2
	push	Acc3

			disp_write                  ; all pins to output, rw 0

	mov		Acc, RP0
	swap    Acc
	andi    Acc, 0x0f                   ; clear Acc high nibble

	in      Acc2, DISP_PORT
	andi    Acc2, 0xf0

	or      Acc2, Acc                   ; Acc2 is the port byte
	out     DISP_PORT, Acc2

	push	RP0
			disp_pulse_enable
	pop		RP0

	in      Acc2, DISP_PORT
	andi    Acc2, 0xf0                  ; clear Acc2 low nibble

	mov		Acc, RP0
	andi    Acc, 0x0f                   ; use Acc low nibble
	or      Acc2, Acc                   ; put it together

	out     DISP_PORT, Acc2                 ; write to display

			disp_pulse_enable

	in      Acc3, DISP_PORT

			disp_instruction
_disp_wait_busy:
	rcall   disp_read_byte

	sbrc	RRet, 7
	rjmp    _disp_wait_busy

	andi    Acc3, 0xf0
	out     DISP_PORT, Acc3                 ; restore instruction / data line

	pop		Acc3
	pop		Acc2
	pop		Acc

	ret

; --------------------------------------------------------------------------
; Proc: Read a byte from the display
;
; Returns: the byte read
; --------------------------------------------------------------------------
disp_read_byte:
	push	Acc

	disp_read
	disp_enable

	ldi		RP0, DISP_SHORT_DELAY
	rcall   timer_start8

	in      RRet, PINB

	disp_disable

	ldi		RP0, DISP_SHORT_DELAY
	rcall   timer_start8

	swap    RRet
	andi    RRet, 0xf0

	disp_enable

	ldi		RP0, DISP_SHORT_DELAY
	rcall   timer_start8

	in      Acc, PINB
	; save Acc before rcall!
	push	Acc

			disp_disable
	ldi		RP0, DISP_SHORT_DELAY
	rcall   timer_start8

	; restore Acc
	pop		Acc

	andi    Acc, 0x0f
	or      RRet, Acc

	pop		Acc

	ret


; --------------------------------------------------------------------------
; Proc: Initialization of the display
;       takes the display to a defined state
; --------------------------------------------------------------------------
disp_reset:
	disp_instruction

	ldi		RP0, 0x28					; 4bit, 2 lines
	rcall   disp_put_byte

	;rcall   disp_long_delay

	ldi		RP0, 0x02					; cursor first pos
	rcall   disp_put_byte

	;rcall   disp_long_delay

	ldi		RP0, 0x01					; clear display
	rcall   disp_put_byte

	;rcall   disp_long_delay

	ldi		RP0, 0x0c					; display on, cursor on (?)
	rcall   disp_put_byte

	ldi		RP0, 0x06					; cursor inc, no shift
	rcall   disp_put_byte

	ret

; --------------------------------------------------------------------------
; Proc: Write a (zero terminated) string to the display
;
; Params: RP0 - low byte, RP1 - high byte of string address in prog mem
; --------------------------------------------------------------------------
disp_write_string:
	push	ZL
	push	ZH
	push	R0

	mov		ZL, RP0
	mov		ZH, RP1

	disp_data
__disp_next_char:
	lpm									; data is in R0 now

	tst		R0
	breq	__disp_write_done

	adiw	ZL, 1

	mov		RP0, R0
	rcall	disp_put_byte

	rjmp	__disp_next_char

__disp_write_done:

	pop		R0
	pop		ZH
	pop		ZL

	ret
